import json
import re
import sys
from pathlib import Path
from subprocess import run

__author__ = "ZeroRin"
log = Path.cwd()/'logs/log.txt'
modified_log =  Path.cwd()/'logs/log.log'

def test_edit_log_path():
    '''
    passing a log_path in string to setup_logging
    all file handlers should be redirected to that file
    '''
    # create config file with multiple file handlers 
    config_file = Path("../logger_tt/log_config.json")
    config = json.loads(config_file.read_text())
    config['handlers']['urllib_file_handler'] = config['handlers']['error_file_handler']
    config['loggers']['urllib3']['handlers'] = ['urllib_file_handler']

    test_config = Path("edit_log_path.json")
    test_config.write_text(json.dumps(config))

    # run test and check outputs
    cmd = [sys.executable, "edit_log_path.py", repr(str(modified_log))]
    run(cmd)
    data = modified_log.read_text()
    assert re.search('A log from default logger', data)
    assert re.search('A log from urllib logger', data)
    test_config.unlink()

def test_edit_log_path_partial():
    '''
    passing a log_path in dict to setup_logging
    only selected file handler should be redirected to that file
    '''
    # create config file with multiple file handlers 
    config_file = Path("../logger_tt/log_config.json")
    config = json.loads(config_file.read_text())
    config['handlers']['urllib_file_handler'] = config['handlers']['error_file_handler']
    config['loggers']['urllib3']['handlers'] = ['urllib_file_handler']
    
    test_config = Path("edit_log_path.json")
    test_config.write_text(json.dumps(config))

    # run test and check outputs
    cmd = [sys.executable, "edit_log_path.py", repr(dict(urllib_file_handler=str(modified_log)))]
    run(cmd)
    data = log.read_text()
    assert re.search('A log from default logger', data)
    data = modified_log.read_text()
    assert re.search('A log from urllib logger', data)
    test_config.unlink()


import re
import sys

import pytest
from ruamel.yaml import YAML
from subprocess import run, PIPE
from pathlib import Path

from logger_tt import setup_logging

__author__ = "Duc Tin"

log = Path.cwd() / 'logs/log.txt'


def test_multiprocessing_normal():
    cmd = [sys.executable, "multiprocessing_normal.py", "3"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert 'Parent process is ready to spawn child' in data
    assert 'child process 0' in data
    assert 'child process 1' in data
    assert 'child process 2' in data


@pytest.mark.parametrize('value', [-1, 2, 'forker', 'spawm'])
def test_multiprocessing_error(value):
    with pytest.raises(ValueError) as e:
        setup_logging(use_multiprocessing=value)

    assert f'Expected a bool or a multiprocessing start_method name, but got: {value}' in str(e)


def test_multiprocessing_pool():
    cmd = [sys.executable, "multiprocessing_pool.py", "10"]
    run(cmd)

    data = log.read_text(encoding='utf8')
    assert 'Parent process pool is ready to spawn child' in data
    assert 'child process 0' in data
    assert 'child process 1' in data
    assert 'child process 2' in data
    assert 'child process 3' in data
    assert 'child process 4' in data
    assert 'child process 5' in data
    assert 'child process 6' in data
    assert 'child process 7' in data
    assert 'child process 8' in data
    assert 'child process 9' in data


def test_multiprocessing_threading():
    """Test a default logger"""
    cmd = [sys.executable, "multiprocessing_threading.py", "10"]
    result = run(cmd, stdout=PIPE, universal_newlines=True)
    assert 'Parent process is ready to spawn child' in result.stdout
    expect = re.findall(r'Process-\d+ Thread-\d+.*? thread running from process', result.stdout)
    assert len(expect) == 10

    data = log.read_text(encoding='utf8')
    assert 'Parent process is ready to spawn child' in data
    expect = re.findall(r'Process-\d+ Thread-\d+.*? thread running from process', data)
    assert len(expect) == 10


def test_multiprocessing_rollover():
    # reduce the rollover interval to 5 seconds
    yaml = YAML(typ='safe')
    config_file = Path("../logger_tt/log_config.yaml")
    log_config = yaml.load(config_file.read_text())
    log_config['handlers']['error_file_handler']['when'] = 's'
    log_config['handlers']['error_file_handler']['interval'] = 5

    # write the config out
    test_config = Path("multiprocessing_issue3_config.yaml")
    yaml.dump(data=log_config, stream=test_config)

    # test it
    cmd = [sys.executable, "multiprocessing_issue3.py", "3"]
    result = run(cmd, stderr=PIPE, universal_newlines=True)
    test_config.unlink()
    assert "PermissionError: [WinError 32] The process cannot access the file" not in result.stderr


def test_multiprocessing_issue5():
    """child processes also create a log path which is unnecessary"""

    # test it
    cmd = [sys.executable, "multiprocessing_issue5.py", "3"]
    result = run(cmd, stdout=PIPE, universal_newlines=True)
    log_parent = log.parent
    sub_folders = [x for x in log_parent.iterdir() if x.is_dir()]
    number_of_path = len(sub_folders)

    # issue 5
    assert number_of_path == 1, sub_folders

    # to be sure
    logfile = sub_folders[0] / 'info.log'
    assert logfile.exists()
    assert 'Test for issue 5' in logfile.read_text()


def test_multiprocessing_port_change():
    """Change the tcp server's port to a user selected one"""
    yaml = YAML(typ='safe')
    config_file = Path("../logger_tt/log_config.yaml")
    log_config = yaml.load(config_file.read_text())
    log_config['logger_tt']['use_multiprocessing'] = True
    log_config['logger_tt']['port'] = 6789

    # write the config out
    test_config = Path("multiprocessing_change_port.yaml")
    yaml.dump(data=log_config, stream=test_config)

    cmd = [sys.executable, "multiprocessing_change_port.py", "3"]
    result = run(cmd, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    assert result.returncode == 0, f'subprocess crashed with error: {result.stderr}'

    test_config.unlink()
    assert '6789' in result.stdout, "Port failed to change"
    assert result.stdout.count("stopped") == 3, "Child process failed to log"




import sys
from logging import getLogger

from logger_tt import setup_logging

__author__ = "ZeroRin"

setup_logging(analyze_raise_statement=True)
logger = getLogger(__name__)

def foo():
    try:
        a=1
        raise RuntimeError(a)
    except:
        try:
            b=2
            raise RuntimeError(b)
        except:
            c=3
            raise RuntimeError(c)

if __name__ == '__main__':
    if sys.argv[-1] == 'caught':
        try:
            foo()
        except:
            logger.exception('Caught exception:')
    elif sys.argv[-1] ==  'uncaught':
        foo()
    else:
        raise ValueError('Usage [PROG] {caught,uncaught}')


import time
import sys
from random import randint

from multiprocessing import Process
from logger_tt import setup_logging
from logging import getLogger


__author__ = "Duc Tin"
logger = getLogger(__name__)
config = setup_logging(config_path="multiprocessing_change_port.yaml")


def worker(arg):
    logger.info(f'child process {arg}: started')
    time.sleep(randint(3, 10))
    logger.info(f'child process {arg}: stopped')


if __name__ == '__main__':
    if len(sys.argv) > 1:
        proc_no = int(sys.argv[1])
    else:
        proc_no = 7

    all_processes = []
    logger.info('Parent process is ready to spawn child')
    logger.info(f'current logging tcp_server is: {config.tcp_server.server_address}')
    for i in range(proc_no):
        p = Process(target=worker, args=(i,))
        all_processes.append(p)
        p.daemon = True
        p.start()

    for p in all_processes:
        p.join()

    print('__finished__')



import datetime
import os
import re
import time

from io import StringIO, BytesIO
from logging import getLogger, DEBUG, Formatter

import pytest
from logger_tt.handlers import StreamHandlerWithBuffer, TelegramHandler


@pytest.mark.parametrize('threshold', [0.2, 0.4])
def test_handler_with_buffer_time(caplog, threshold):
    logger = getLogger('Test buffer time')
    my_stream = StringIO()
    handler = StreamHandlerWithBuffer(stream=my_stream, buffer_time=threshold, buffer_lines=0, debug=True)
    formatter = Formatter(fmt="[%(asctime)s.%(msecs)03d] %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False

    with caplog.at_level(DEBUG):
        t0 = time.time()
        dt = 0
        while dt < 2*threshold + 2*0.016:
            logger.info('This is my log message')
            time.sleep(0.01)
            dt = time.time() - t0

        my_stream.seek(0)
        logs = my_stream.read()

        for i in range(2):
            time.sleep(0.2)
            data = re.findall(r'StreamHandlerWithBuffer.*\n', logs)
            try:
                start_time = datetime.datetime.strptime(data[0].strip().split(': ')[1], '%Y-%m-%d %H:%M:%S.%f')
                flush1_time = datetime.datetime.strptime(data[1].strip().split(': ')[1], '%Y-%m-%d %H:%M:%S.%f')
                flush2_time = datetime.datetime.strptime(data[2].strip().split(': ')[1], '%Y-%m-%d %H:%M:%S.%f')
                break
            except IndexError:
                continue
        else:
            raise IndexError

        dt1 = (flush1_time - start_time).microseconds/1e6
        dt2 = (flush2_time - flush1_time).microseconds/1e6
        assert threshold < dt1 < threshold + 0.2*threshold, "It shouldn't flush too soon or too late"
        assert threshold < dt2 < threshold + 0.2*threshold, "It shouldn't flush too soon or too late"

        lines = logs.splitlines()
        msg_count = lines.index(data[2].strip()) - lines.index(data[1].strip()) - 1
        assert msg_count, "Some messages must be logged between 2 flushing"


@pytest.mark.parametrize('threshold', [10, 20])
def test_handler_with_buffer_lines(caplog, threshold):
    logger = getLogger('Test buffer lines')
    my_stream = StringIO()
    handler = StreamHandlerWithBuffer(stream=my_stream, buffer_time=0, buffer_lines=threshold)
    formatter = Formatter(fmt="[%(asctime)s.%(msecs)03d] %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False

    with caplog.at_level(DEBUG):
        for i in range(50):
            logger.info('This is my log line')
            if i == threshold - 2:
                my_stream.seek(0)
                logs = my_stream.read()
                assert not logs, "Right before threshold exceeding, no log should appear"
            if i == threshold - 1:
                my_stream.seek(0)
                logs = my_stream.read()
                assert logs, "Right after threshold exceeding, the log should appear"
            if i == threshold*2-1:
                my_stream.seek(0)
                logs = my_stream.read().splitlines()
                assert len(logs) == 2*threshold, "Total log lines should equal threshold * 2"


@pytest.mark.skip('Uncomment this line. Add your bot token and chat_id/(group_id, topic) to test. Already tested!')
@pytest.mark.parametrize('unique_id', ['123456789',
                                       '-1234567890123@2',
                                       '-1234567890123@2; -1234567890123@4'
                                       ])
def test_telegram_handler_basic(unique_id):
    bot_token = 'your bot token here'
    os.environ['TELEGRAM_BOT_LOG_TOKEN'] = bot_token
    os.environ['TELEGRAM_BOT_LOG_DEST'] = unique_id
    logger = getLogger('test telegram')
    handler = TelegramHandler(env_token_key='TELEGRAM_BOT_LOG_TOKEN', env_unique_ids_key='TELEGRAM_BOT_LOG_DEST')
    formatter = Formatter(fmt="[%(asctime)s.%(msecs)03d] %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False

    try:
        logger.warning('this is my warning')
        feedback = handler.feedback[unique_id]
        assert feedback.get('ok')

        logger.error('this is my error \n this is the next line')
        feedback = handler.feedback[unique_id]
        assert feedback.get('ok')
    finally:
        # remove handler to test a second params set
        logger.removeHandler(handler)
        handler.close()


def test_telegram_handler_error(caplog):
    bot_token = ''
    user_id = '123456789'
    logger = getLogger('test telegram 0')
    handler = TelegramHandler(token=bot_token, unique_ids=user_id, debug=True, check_interval=0.5)
    formatter = Formatter(fmt="[%(asctime)s.%(msecs)03d] %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False
    getLogger().setLevel(0)

    # setup stub function
    from urllib import request, error

    def fake_urlopen(*arg, **kwargs):
        if code != 200:
            raise error.HTTPError('fake_url', code=code, msg='fake_url raised error', hdrs={}, fp=None)
        else:
            my_stream = BytesIO()
            my_stream.write(b'{"ok": "true"}')
            my_stream.seek(0)
            return my_stream

    request.urlopen = fake_urlopen
    # stub done

    code = 403
    logger.warning('user blocked this bot')
    assert 'HTTP Error 403' in caplog.text
    assert not handler.cache[user_id]

    code = 500
    logger.error('server error')
    assert handler.cache[user_id]
    code = 200

    for retry in range(2):
        try:
            time.sleep(1.5)
            assert not handler.cache[user_id]
            break
        except AssertionError:
            continue
    else:
        assert not handler.cache[user_id]
    assert 'HTTP Error 500' in caplog.text
    assert 'found unsent messages' in caplog.text


def test_telegram_handler_repeated_msg_continuous(caplog):
    bot_token = ''
    user_id = '123456789'
    logger = getLogger('test telegram 1')
    handler = TelegramHandler(token=bot_token, unique_ids=user_id, debug=True, check_interval=1)
    formatter = Formatter(fmt="[%(asctime)s.%(msecs)03d] %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False
    getLogger().setLevel(0)

    # setup
    from urllib import request, error
    code = 200
    log_sent = StringIO()

    def fake_urlopen(url: str, *args, **kwargs):
        if code != 200:
            raise error.HTTPError('fake_url', code=code, msg='fake_url raised error', hdrs={}, fp=None)
        else:
            log_sent.write(url+'\n\n')
            my_stream = BytesIO()
            my_stream.write('{"ok": "true"}\n'.encode())
            my_stream.seek(0)
            return my_stream

    request.urlopen = fake_urlopen

    # run repeat the same message continuously
    for i in range(1000):
        logger.warning(f'Connection error: server 500. Retry')
    time.sleep(2)   # let watcher run

    # check result
    log_sent.seek(0)
    data = log_sent.read()
    count = data.count('Connection+error')
    assert count < 500
    res = re.findall(r'Message\+repeated\+(\d+)\+times', data)
    assert sum(int(x) for x in res) == 1000 - 1, data + '\n\n' + caplog.text


def test_telegram_handler_repeated_msg_then_change(caplog):
    bot_token = ''
    user_id = '123456789'
    logger = getLogger('test telegram 2')
    handler = TelegramHandler(token=bot_token, unique_ids=user_id, debug=True, check_interval=1)
    formatter = Formatter(fmt="[%(asctime)s.%(msecs)03d] %(levelname)s: %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False
    getLogger().setLevel(0)

    # setup
    from urllib import request, error
    code = 200
    log_sent = StringIO()

    def fake_urlopen(url: str, *args, **kwargs):
        if code != 200:
            raise error.HTTPError('fake_url', code=code, msg='fake_url raised error', hdrs={}, fp=None)
        else:
            log_sent.write(url+'\n\n')
            my_stream = BytesIO()
            my_stream.write('{"ok": "true"}\n'.encode())
            my_stream.seek(0)
            return my_stream

    request.urlopen = fake_urlopen

    # run repeat the same message for a while then different message
    log_sent = StringIO()
    for i in range(10):
        logger.warning(f'Connection error: server 500. Retry')
    else:
        logger.error(f'Failure. Memory overflow')
    time.sleep(2)   # let watcher run

    # check result, watcher should has nothing to do
    log_sent.seek(0)
    data = log_sent.read()
    count = data.count('Connection+error')
    assert 1 < count < 6, data + '\n\n' + caplog.text
    res = re.findall(r'Message\+repeated\+(\d+)\+times', data)
    assert 'Message+repeated+9+times' in data
    assert sum(int(x) for x in res) == 10 - len(res), data + '\n\n' + caplog.text
    assert 'Memory+overflow' in data
    assert 'watcher emit duplicated' not in caplog.text



import logging

from logging import getLogger
from .inspector import logging_disabled
from .core import LogConfig


__author__ = "Duc Tin"
__all__ = ['setup_logging', 'logging_disabled', 'getLogger', 'logger']


logger: logging.Logger


def setup_logging(config_path: str = "", log_path: str = "",
                  capture_print: bool = False,
                  strict: bool = False,
                  guess_level: bool = False,
                  full_context: int = 0,
                  suppress: list = None,
                  suppress_level_below: int = logging.WARNING,
                  use_multiprocessing: bool = False,
                  limit_line_length: int = 1000,
                  analyze_raise_statement: bool = False,
                  host: str = None,
                  port: int = None) -> LogConfig: ...



import sys
from logger_tt import setup_logging, logger


def will_fail():
    loren_ipsum = "On the other hand, we denounce with righteous indignation and dislike men who are so beguiled and " \
                  "demoralized by the charms of pleasure of the moment, so blinded by desire, that they cannot " \
                  "foresee the pain and trouble that are bound to ensue; and equal blame belongs to those who fail in " \
                  "their duty through weakness of will, which is the same as saying through shrinking from toil and " \
                  "pain. These cases are perfectly simple and easy to distinguish. In a free hour, when our power of " \
                  "choice is untrammelled and when nothing prevents our being able to do what we like best, " \
                  "every pleasure is to be welcomed and every pain avoided. But in certain circumstances and owing to " \
                  "the claims of duty or the obligations of business it will frequently occur that pleasures have to " \
                  "be repudiated and annoyances accepted. The wise man therefore always holds in these matters to " \
                  "this principle of selection: he rejects pleasures to secure other greater pleasures, or else he " \
                  "endures pains to avoid worse pains. "

    raise RuntimeError(f'Below is the random text used as a standard to test font: \n{loren_ipsum}')


if __name__ == '__main__':
    limit_line_length = int(sys.argv[1])
    analyze_raise = bool(int(sys.argv[2]))
    setup_logging(limit_line_length=limit_line_length, analyze_raise_statement=analyze_raise)
    logger.info(sys.argv)
    will_fail()



import re
import sys
from pathlib import Path
from subprocess import run

__author__ = "ZeroRin"
log = Path.cwd() / 'logs/log.txt'

TRACE=r'''Traceback \(most recent call last\):
  File ".+?", line \d+?, in foo
    raise RuntimeError\(a\)
     |-> a = 1
RuntimeError: 1

During handling of the above exception, another exception occurred:

Traceback \(most recent call last\):
  File ".+?", line \d+, in foo
    raise RuntimeError\(b\)
     |-> b = 2
RuntimeError: 2

During handling of the above exception, another exception occurred:

Traceback \(most recent call last\):
  File ".+", line \d+, in <module>
    foo\(\)

  File ".+", line \d+, in foo
    raise RuntimeError\(c\)
     |-> c = 3
RuntimeError: 3'''

def test_recur_exception_caught():
    cmd = [sys.executable, "exception_on_exception.py", "caught"]
    run(cmd)
    data = log.read_text()
    pattern = rf'Caught exception\n{TRACE}'
    assert re.search(pattern, data, re.DOTALL)

def test_recur_exception_uncaught():
    cmd = [sys.executable, "exception_on_exception.py", "uncaught"]
    run(cmd)
    data = log.read_text()
    pattern = rf'Uncaught exception\n{TRACE}'
    assert re.search(pattern, data, re.DOTALL)


import logging
import time
import os
import json
from urllib import request, parse, error
from collections import deque
from threading import Thread, Event
from datetime import datetime


class StreamHandlerWithBuffer(logging.StreamHandler):
    def __init__(self, stream=None, buffer_time: float = 0.2, buffer_lines: int = 50, debug=False):
        super().__init__(stream)
        assert buffer_time >= 0 or buffer_lines >= 0, "At least one kind of buffer must be set"

        self.buffer_time = buffer_time
        self.buffer_lines = buffer_lines
        self.buffer = []
        self.debug = debug

        self._stop_event = Event()
        if self.buffer_time:
            watcher = Thread(target=self.watcher, daemon=True)
            watcher.start()

    def close(self) -> None:
        self._stop_event.set()

    def export(self):
        """Actual writing data out to the stream"""

        if self.debug:
            self.buffer.append(f'StreamHandlerWithBuffer flush: {datetime.now()}')

        msg = self.terminator.join(self.buffer)
        stream = self.stream
        # issue 35046: merged two stream.writes into one.
        stream.write(msg + self.terminator)
        self.flush()

        self.buffer.clear()

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            msg = self.format(record)
            self.buffer.append(msg)
            if self.buffer_lines and len(self.buffer) >= self.buffer_lines:
                self.export()

        except RecursionError:  # See issue 36272
            raise
        except Exception:
            self.handleError(record)

    def watcher(self):
        """
        If buffer_time is used, this method will flush the buffer
        after every buffer_time seconds has passed.
        """
        if self.debug:
            self.buffer.append(f'StreamHandlerWithBuffer watcher starts: {datetime.now()}')
        while not self._stop_event.is_set():
            time.sleep(self.buffer_time)
            if self.buffer:
                self.acquire()
                self.export()
                self.release()


class TelegramHandler(logging.Handler):
    def __init__(self, token='', unique_ids=None, env_token_key='', env_unique_ids_key='',
                 debug=False, check_interval=300):
        super().__init__()

        if env_token_key:
            token = os.environ.get(env_token_key, None) or token
        if env_unique_ids_key:
            unique_ids = os.environ.get(env_unique_ids_key, None) or unique_ids

        self._unique_ids = []       # type: list[str]
        self.set_unique_ids(unique_ids)

        self._url = f"https://api.telegram.org/bot{token}/sendMessage"
        self.feedback = {x: {} for x in self._unique_ids}
        self.cache = {x: deque(maxlen=100) for x in self._unique_ids}

        # back ground thread resends the log if network error previously
        self.debug = debug
        self.check_interval = check_interval
        self._stop_event = Event()
        watcher_thread = Thread(target=self.watcher, daemon=True)
        watcher_thread.start()

        # reduce sending duplicated log
        self.last_record = None
        self.dup_count = 0

    def close(self) -> None:
        self._stop_event.set()

    def _get_full_url(self, unique_id: str, text: str):
        # remove name/label if presence
        unique_id = unique_id.split(':')[-1]

        # add message_thread_id if presence
        if '@' in unique_id:
            # group_id and topic index is specified
            chat_id, message_thread_id = unique_id.split('@')
            url = f'{self._url}?chat_id={chat_id}&message_thread_id={message_thread_id}&text={text}'
        else:
            # just chat_id only
            url = f'{self._url}?chat_id={unique_id}&text={text}'

        return url

    def set_bot_token(self, token: str):
        self._url = f"https://api.telegram.org/bot{token}/sendMessage"

    def set_unique_ids(self, ids):
        if not ids:
            self._unique_ids = []
        elif type(ids) is str:
            # str from environment variable
            self._unique_ids = [x.strip() for x in ids.split(';')]
        elif type(ids) is int:
            # from config, one value
            self._unique_ids = [str(ids)]
        else:
            raise TypeError(f'Expected str or int but got type: {type(ids)}')

    def send(self):
        for _id_ in self._unique_ids:
            while self.cache[_id_]:
                msg_out = self.cache[_id_][0]
                full_url = self._get_full_url(_id_, msg_out)

                try:
                    with request.urlopen(full_url) as fi:
                        data = fi.read()
                    self.feedback[_id_] = json.loads(data.decode())

                    # remove from the queue after sending successfully
                    self.cache[_id_].popleft()

                except json.JSONDecodeError as e:
                    self.feedback[_id_] = {'error': str(e), 'data': data}

                except error.HTTPError as e:
                    if e.code == 403:
                        # user blocked the bot
                        logging.getLogger().error(e)

                        # remove msg
                        self.cache[_id_].popleft()
                        break
                    else:
                        # resend this msg later
                        logging.getLogger().error(e)
                        break

                except Exception as e:
                    # resend this later
                    logging.getLogger().exception(e)
                    break

    def _is_duplicated_record(self, record):
        if not self.last_record:
            self.last_record = record
            self.dup_count = 0
            return False

        for attr in 'msg name levelno pathname lineno args funcName'.split():
            if getattr(record, attr) != getattr(self.last_record, attr):
                return False
        else:
            return True

    def _naive_emit(self, record, remark=''):
        # cache msg in case of sending failure
        msg = self.format(record) + remark

        # redirect msg to appropriate cache
        if getattr(record, 'dest_name', ''):
            dest_id = next(filter(lambda x: x.startswith(f'{record.dest_name}:'), self._unique_ids), None)
            if dest_id:
                self.cache[dest_id].append(parse.quote_plus(msg))
            else:
                # do nothing
                pass
        else:
            for _id_ in self._unique_ids:
                self.cache[_id_].append(parse.quote_plus(msg))
        self.send()

    def emit(self, record):
        self.acquire()

        if self._is_duplicated_record(record):
            self.dup_count += 1

        elif self.dup_count:
            # changed to new record, no longer duplicated
            # send last msg, then send this time msg
            self._naive_emit(self.last_record, remark=f'\n (Message repeated {self.dup_count} times)')
            self._naive_emit(record)

            self.last_record = record
            self.dup_count = 0
        else:
            # last sent record is not duplicated
            self._naive_emit(record)
            self.last_record = record

        self.release()

    def watcher(self):
        """
        This method will resend the failed messages if they haven't been sent in emit
        """
        if self.debug:
            logging.getLogger().debug(f'TelegramHandler watcher starts: {datetime.now()}')

        while not self._stop_event.is_set():
            time.sleep(self.check_interval)
            if any(self.cache.values()):
                if self.debug:
                    logging.getLogger().debug(f'TelegramHandler found unsent messages: {datetime.now()}')
                self.acquire()
                self.send()
                self.release()
            elif self.dup_count > 1:
                if self.debug:
                    logging.getLogger().debug(f'TelegramHandler watcher emit duplicated msg with: {datetime.now()}')
                self.acquire()
                self._naive_emit(self.last_record, f'\n (Message repeated {self.dup_count} times)')
                self.dup_count = 0
                self.release()



from logger_tt import setup_logging
from logging import getLogger

__author__ = "Duc Tin"
setup_logging()
logger = getLogger(__name__)


def foo(*args, **kwargs):
    return 1


def lel(x, e=4):
    a = 1
    b = 2
    for i in range(1):
        return 1 + foo(a, b, c=x['z']['x']['y']['z']['y'], d=e)


def lel2(x):
    return 25 + lel(x) + lel(x)


def lel3(x):
    return lel2(x) / 23


if __name__ == '__main__':
    xx = {'z': {'x': {'y': None}}}
    lel3(xx)



__author__ = "Duc Tin"

import re
import logging
import tokenize
import linecache
from io import StringIO
from contextlib import contextmanager

ID_PATTERN = re.compile(r'([a-zA-Z_][a-zA-Z_0-9.]*)')
MEM_PATTERN = re.compile(r'<.*object at 0x[0-9A-Fa-f]+>')


try:
    # python 3.11.0, traceback indicator
    from traceback import StackSummary, _walk_tb_with_full_positions as walk_tb
    stack_summary_extract = StackSummary._extract_from_extended_frame_gen
except ImportError as e:
    from traceback import StackSummary, walk_tb
    stack_summary_extract = StackSummary.extract


def get_recur_attr(obj, attr: str):
    """
    Follow the dot `.` in attribute string `attr` to the final object

    :param obj: any object
    :param attr: string of attribute access
    :return: the final desire object or '!!! Not Exists'

    example: if we need to access `c` object as in `a.b.c`,
    then `obj=a, attr='b.c'`
    If `b` or `c` doesn't exists, then `'!!! Not Exists'` is returned
    """
    this_level, *next_levels = attr.split('.', maxsplit=1)
    if not next_levels:
        try:
            return getattr(obj, this_level)
        except AttributeError:
            return "!!! Not Exists"
        except Exception as e:
            return f"!!! Attribute Error: {e}"
    else:
        try:
            obj = getattr(obj, this_level)
        except AttributeError:
            return "!!! Not Exists"
        return get_recur_attr(obj, next_levels[0])


def get_repr(obj, multiline_indent: int = 0) -> str:
    """
    Pick a more useful representation of object `obj` between __str__ and __repr__
     if it available.

    :param obj: any of object
    :param multiline_indent: indent level for the second line onward
    :return: string of object representation
    """
    try:
        _repr_ = repr(obj)
        if not MEM_PATTERN.match(_repr_):
            return _repr_
    except Exception as e:
        _repr_ = f"!!! repr error: {e}"

    try:
        _str_ = str(obj).replace('\r\n', '\n')
        if '\n' in _str_:
            indent = ' ' * multiline_indent
            _str_ = _str_.replace('\n', '\n' + indent)

        return _str_

    except Exception as e:
        return _repr_


def is_half_ended(line: str) -> bool:
    """
    Check if we a given a last line of a multiline statement with brackets
    """
    if line.count(')') > line.count('('):
        return True
    if line.count('}') > line.count('{'):
        return True
    if line.count(']') > line.count('['):
        return True


def is_full_statement(*lines: str) -> bool:
    """
    Check if a set of lines makes up a full python statement

    :param lines: list of line of python code
    :return: True if line in `lines` made up a full python statement
    """
    try:
        stream = StringIO()
        stream.write('\n'.join(lines))
        stream.seek(0)
        for t in tokenize.generate_tokens(stream.readline):
            pass
    except tokenize.TokenError as e:
        if 'EOF in multi-line' in str(e):
            return False
        else:
            return True
    else:
        return True


def get_statement_up(filename: str, lineno: int) -> list:
    """The python interpreter return the last line of exception lines
        This behavior is of python 3.6, 3.7
        We have to go up to grab the full statement

        Grab maximum 10 lines
    """
    startno = lineno
    lines = []

    while lineno >= max(0, startno - 9):
        line = linecache.getline(filename, lineno)
        lines.insert(0, line)

        temp = ''.join(lines)
        if not is_half_ended(temp):
            break

        lineno -= 1

    return lines


def get_statement_down(filename: str, lineno: int) -> list:
    """The python interpreter return the first line of exception lines
        This behavior is of python 3.9
        We go down normally to grab the full statement

        Grab maximum 10 lines
    """
    lines = []
    startno = lineno
    while lineno <= startno + 9:
        line = linecache.getline(filename, lineno)
        if not line:
            break

        lines.append(line)
        if is_full_statement(*lines):
            break

        lineno += 1

    return lines


def get_full_statement(filename, lineno: int) -> list:
    """
    Get all lines of python file `filename` that makes up a full statement starting from `lineno`

    :param filename: path to python source code file
    :param lineno: line number at with the statement started
    :return: list of maximum 5 lines that made up a python statement with indent striped,
    """

    line = linecache.getline(filename, lineno)
    if is_half_ended(line):
        lines = get_statement_up(filename, lineno)
    else:
        lines = get_statement_down(filename, lineno)

    lines[-1] = lines[-1].strip('\n')  # remove newline in last line
    indent = re.search(r'^(\s+)', lines[0])
    if indent:
        indent = indent.group(1)
        return [x.replace(indent, '', 1) for x in lines]
    else:
        return lines


@contextmanager
def logging_disabled():
    logging.disable(logging.CRITICAL)
    yield
    logging.disable(logging.NOTSET)


def get_traceback_depth(trace_back) -> int:
    count = 0
    tb = trace_back
    while tb is not None:
        tb = tb.tb_next
        count += 1

    return count


def get_basic_exception_info(summary) -> tuple:
    """Return a normal 2 lines of of an exception and a full python statement in case of multiline"""

    # no indicator
    line = summary.line
    if not is_full_statement(summary.line):
        line = get_full_statement(summary.filename, summary.lineno)
        line = '    '.join(line)

    if not getattr(summary, 'colno', None):
        txt = [f'  File "{summary.filename}", line {summary.lineno}, in {summary.name}',
               f'    {line}']
    else:
        # indicator
        _locals = summary.locals
        summary.locals = {}
        data = StackSummary.format_frame_summary(None, summary)
        txt = data.splitlines()
        summary.locals = _locals

    return txt, line


def parse_line(identifiers, frame, outer) -> list:
    """List variables appear on the exception line and their values"""
    seen = set()

    bullet_1 = '|->'
    multi_line_indent1 = 8 + len(bullet_1)

    global_var = frame.f_globals
    local_var = frame.f_locals

    txt = []
    for i in identifiers:
        if i in seen or i.endswith('.'):
            continue

        seen.add(i)
        spaces = multi_line_indent1 + len(i)
        if i in local_var:
            value = get_repr(local_var[i], spaces)
            txt.append(f'     {bullet_1} {i} = {value}')
        elif i in global_var:
            spaces += len(outer)
            value = get_repr(global_var[i], spaces)
            txt.append(f'     {bullet_1} {outer}{i} = {value}')
        elif '.' in i:
            # class attribute access
            spaces += len(outer)
            instance = i.split('.')[0]
            obj = local_var.get(instance, global_var.get(instance))
            attribute = get_recur_attr(obj, i[len(instance) + 1:])
            value = get_repr(attribute, spaces)
            scope = outer if instance in global_var else ''
            txt.append(f'     {bullet_1} {scope}{i} = {value}')
        else:
            # reserved Keyword or non-identifier, eg. word inside the string
            pass

    return txt


def parse_full_context(identifiers, frame) -> list:
    """List variables within the scope of the exception line and their values"""
    bullet_2 = '=>'
    multi_line_indent2 = 8 + len(bullet_2)

    local_var = frame.f_locals
    other_local_var = set(local_var) - set(identifiers)
    txt = []
    if other_local_var:
        spaces = multi_line_indent2
        txt = [f'     {bullet_2} {k} = {get_repr(v, spaces + len(k))}'
               for k, v in local_var.items() if k in other_local_var]
    return txt


def analyze_exception_recur(e: BaseException, full_context: int, limit_line_length: int, analyze_raise_statement: bool,
                            text: str = '') -> str:
    cause = e.__cause__ or e.__context__
    if cause:
        text = analyze_exception_recur(cause, full_context, limit_line_length, analyze_raise_statement, text)
        text += '\nDuring handling of the above exception, another exception occurred:\n\n'

    text += ("Traceback (most recent call last):\n"
             + analyze_frame(e.__traceback__, full_context, limit_line_length, analyze_raise_statement)
             + f'{type(e).__name__}: {e}\n')
    return text


def analyze_frame(trace_back, full_context: int, limit_line_length: int, analyze_raise_statement: bool) -> str:
    """
    Read out variables' content surrounding the error line of code

    :param trace_back: A traceback object when exception occur
    :param full_context: Also export local variables that is not in the error line
            full_context == 0: export only variables that appear on the error line
            full_context == 1: export variables within the error function's scope
            full_context >= 2: export variables along the function's call stack up to `full_context` level
    :param limit_line_length: maximum character on one line of traceback, 0 for unlimited
    :param analyze_raise_statement: should the variables in `raise` exception line be shown or not
    :return: string of analyzed frame
    """
    result = []
    full_context = max(0, int(full_context))
    stack_depth = get_traceback_depth(trace_back)
    # todo: add color

    with logging_disabled():
        for idx, obj in enumerate(walk_tb(trace_back)):
            frame, _ = obj

            summary = stack_summary_extract([obj], capture_locals=True)[0]
            txt, line = get_basic_exception_info(summary)

            # todo: dump all level to different file?
            parse_level = max(full_context - 1, 0)
            if idx + 1 < (stack_depth - parse_level):
                # don't parse variables for top levels
                txt.append('')
                result.append('\n'.join(txt))
                continue

            # get value of variables on the error line
            identifiers = ID_PATTERN.findall(line)
            if analyze_raise_statement or not line.strip().startswith('raise '):
                outer = "(outer) " if idx else ""  # ground level variables are not outer for sure
                new_info = parse_line(identifiers, frame, outer)
                txt.extend(new_info)

            # get value of other variables within local scope
            if full_context:
                new_info = parse_full_context(identifiers, frame)
                txt.extend(new_info)

            # limit the length of the line
            if limit_line_length > 0:
                for i, line in enumerate(txt):
                    if len(line) > limit_line_length:
                        char_left = len(line) - limit_line_length
                        txt[i] = line[:limit_line_length] + f'... ({char_left} characters more)'

            txt.append('')
            result.append('\n'.join(txt))

    return '\n'.join(result)



import sys
from logging import getLogger
from typing import *

from logger_tt import logger, setup_logging

__author__ = "ZeroRin"

def modify_log_files(log_path: Union[str,dict]=''):
    print(log_path)
    setup_logging(config_path='edit_log_path.json',log_path=log_path)

    urllib_logger = getLogger('urllib3')
    logger.info('A log from default logger')
    urllib_logger.error('A log from urllib logger')


if __name__=='__main__':
    try:
        log_path = eval(sys.argv[-1])
    except:
        raise ValueError('Usage [PROG] log_path_repr')
    modify_log_files(log_path)